use crate::{builder::ServerBuilder, config::ServerConfig};
use http::HeaderValue;
use http_body_util::Full;
use hyper::{body::Incoming, server::conn::http1, service::service_fn, Request, Response};
use hyper_util::rt::{TokioIo, TokioTimer};
use log::error;
use std::{
    convert::Infallible,
    future::Future,
    net::{IpAddr, SocketAddr},
    sync::Arc,
};
use tokio::{net::TcpListener, spawn};
use tokio_util::bytes::Bytes;

pub struct Server {
    pub config: Arc<ServerConfig>,
}

impl Server {
    pub fn builder() -> ServerBuilder {
        ServerBuilder {
            config: ServerConfig::default(),
        }
    }

    pub async fn run<F, Fut>(self, handler: F)
    where
        F: Fn(Request<Incoming>) -> Fut + Send + Sync + 'static,
        Fut: Future<Output = Result<Response<Full<Bytes>>, Infallible>> + Send,
    {
        let addr: SocketAddr = format!("{}:{}", self.config.ip, self.config.port)
            .parse()
            .expect("Invalid IP or port");
        let listener = TcpListener::bind(addr).await.unwrap();
        let handler = Arc::new(handler);

        loop {
            let (tcp, client_addr) = listener.accept().await.unwrap();
            let io = TokioIo::new(tcp);

            let config = Arc::clone(&self.config);
            if self.config.ips_filter && !self.is_ip_authorized(&client_addr.ip()) {
                continue;
            }

            let handler = Arc::clone(&handler);
            spawn(async move {
                if let Err(error) = http1::Builder::new()
                    .timer(TokioTimer::new())
                    .serve_connection(
                        io,
                        service_fn(move |req| {
                            let config = Arc::clone(&config);
                            let handler = Arc::clone(&handler);

                            async move {
                                if let Some(ref token) = config.api_key {
                                    if req.headers().get("X-API-Key")
                                        != Some(&HeaderValue::from_str(token).unwrap())
                                    {
                                        return Ok(Response::builder()
                                            .status(401)
                                            .body(Full::new(Bytes::from("Unauthorized")))
                                            .unwrap());
                                    }
                                }

                                handler(req).await
                            }
                        }),
                    )
                    .await
                {
                    error!(error = error.to_string().as_str();
                        "Serving connection"
                    );
                }
            });
        }
    }

    fn is_ip_authorized(&self, ip: &IpAddr) -> bool {
        if self.config.private_ips {
            let is_private = match ip {
                IpAddr::V4(ip4) => ip4.is_private(),
                IpAddr::V6(_) => false,
            };

            if is_private {
                return true;
            }
        }

        let ips = &self.config.ips;
        if ips.is_empty() {
            return true;
        }

        ips.iter()
            .any(|authorized_ip| &ip.to_string() == authorized_ip)
    }
}
